@@ -1,20 +1,26 @@
 /* gifread.c - Functions to read GIFs.
-   Copyright (C) 1997 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of the GIF library.
 
-   The GIF library is free software*; you can copy, distribute, or alter it at
-   will, as long as this notice is kept intact and this source code is made
-   available. Hypo(pa)thetical commerical developers are asked to write the
-   author a note, which might make his day. There is no warranty, express or
-   implied.
+   The GIF library is free software*. It is distributed under the GNU Public
+   License, version 2 or later; you can copy, distribute, or alter it at will,
+   as long as this notice is kept intact and this source code is made
+   available. There is no warranty, express or implied.
 
    *The LZW compression method used by GIFs is patented. Unisys, the patent
    holder, allows the compression algorithm to be used without a license in
    software distributed at no cost to the user. */
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#elif !defined(__cplusplus)
+/* Assume we don't have inline by default */
+# define inline
+#endif
 #include "gif.h"
 #include <stdarg.h>
 #include <assert.h>
+#include <string.h>
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -34,6 +40,9 @@ typedef struct {
   byte *maximage;
   
   unsigned decodepos;
+
+  Gif_ReadErrorHandler handler;
+  void *handler_thunk;
   
 } Gif_Context;
 
@@ -44,6 +53,7 @@ typedef struct Gif_Reader {
   byte *v;
   u_int32_t w;
   int is_record;
+  int is_eoi;
   byte (*byte_getter)(struct Gif_Reader *);
   void (*block_getter)(byte *, u_int32_t, struct Gif_Reader *);
   int (*eofer)(struct Gif_Reader *);
@@ -56,10 +66,7 @@ typedef struct Gif_Reader {
 #define gifgetblock(ptr, size, grr) ((*grr->block_getter)(ptr, size, grr))
 #define gifeof(grr)	((*grr->eofer)(grr))
 
-#ifdef __GNUC__
-__inline__
-#endif
-static u_int16_t
+static inline u_int16_t
 gifgetunsigned(Gif_Reader *grr)
 {
   byte one = gifgetbyte(grr);
@@ -121,18 +128,13 @@ make_data_reader(Gif_Reader *grr, byte *data, u_int32_t length)
 }
 
 
-#ifdef GIF_DEBUGGING
-#include <stdarg.h>
-void
-Gif_Debug(char *format, ...)
+static void
+gif_read_error(Gif_Context *gfc, const char *error)
 {
-  va_list val;
-  va_start(val, format);
-  vfprintf(stderr, format, val);
-  putc(' ', stderr);
-  va_end(val);
+  gfc->stream->errors++;
+  if (gfc->handler)
+    gfc->handler(error, gfc->stream->nimages, gfc->handler_thunk);
 }
-#endif
 
 
 static byte
@@ -143,13 +145,13 @@ one_code(Gif_Context *gfc, Gif_Code code)
   byte *ptr;
   int lastsuffix;
   u_int16_t codelength = gfc->length[code];
-  
+
   gfc->decodepos += codelength;
   ptr = gfc->image + gfc->decodepos;
   if (ptr > gfc->maximage || !codelength) {
-    gfc->stream->errors++;
+    gif_read_error(gfc, (!codelength ? "bad code" : "too much image data"));
     /* 5/26/98 It's not good enough simply to count an error, because in the
-       image_data function, if code == next_code, we will store a byte in
+       read_image_data function, if code == next_code, we will store a byte in
        gfc->image[gfc->decodepos-1]. Thus, fix decodepos so it's w/in the
        image. */
     gfc->decodepos = gfc->maximage - gfc->image;
@@ -168,194 +170,210 @@ one_code(Gif_Context *gfc, Gif_Code code)
   return lastsuffix;
 }
 
+static int
+read_image_block(Gif_Reader *grr, byte *buffer, int *bit_pos_store,
+		 int *bit_len_store, int bits_needed)
+{
+  int bit_position = *bit_pos_store;
+  int bit_length = *bit_len_store;
+  byte block_len;
+  
+  while (bit_position + bits_needed > bit_length) {
+    /* Read in the next data block. */
+    if (bit_position >= 8) {
+      /* Need to shift down the upper, unused part of `buffer' */
+      int i = bit_position / 8;
+      buffer[0] = buffer[i];
+      buffer[1] = buffer[i+1];
+      bit_position -= i * 8;
+      bit_length -= i * 8;
+    }
+    block_len = gifgetbyte(grr);
+    GIF_DEBUG(("image_block(%d)", block_len));
+    if (block_len == 0) return 0;
+    gifgetblock(buffer + bit_length / 8, block_len, grr);
+    bit_length += block_len * 8;
+  }
+  
+  *bit_pos_store = bit_position;
+  *bit_len_store = bit_length;
+  return 1;
+}
+
 
 static void
-image_data(Gif_Context *gfc, Gif_Reader *grr)
+read_image_data(Gif_Context *gfc, Gif_Reader *grr)
 {
-  byte buffer[GIF_MAX_BLOCK];
-  byte *buf;
-  byte *bufmax;
+  /* we need a bit more than GIF_MAX_BLOCK in case a single code is split
+     across blocks */
+  byte buffer[GIF_MAX_BLOCK + 5];
   int i;
+  u_int32_t accum;
   
-  int leftover;
-  int amountleftover;
+  int bit_position;
+  int bit_length;
   
   Gif_Code code;
   Gif_Code old_code;
   Gif_Code clear_code;
   Gif_Code eoi_code;
   Gif_Code next_code;
-  Gif_Code max_code;
+#define CUR_BUMP_CODE (1 << bits_needed)
+#define CUR_CODE_MASK ((1 << bits_needed) - 1)
   
   int min_code_size;
   int bits_needed;
-  Gif_Code codemask;
   
   gfc->decodepos = 0;
   
   min_code_size = gifgetbyte(grr);
-  if (min_code_size >= GIF_MAX_CODE_SIZE) {
-    gfc->stream->errors++;
-    min_code_size = GIF_MAX_CODE_SIZE - 1;
+  GIF_DEBUG(("min_code_size(%d)", min_code_size));
+  if (min_code_size >= GIF_MAX_CODE_BITS) {
+    gif_read_error(gfc, "min_code_size too big");
+    min_code_size = GIF_MAX_CODE_BITS - 1;
+  } else if (min_code_size < 2) {
+    gif_read_error(gfc, "min_code_size too small");
+    min_code_size = 2;
   }
   clear_code = 1 << min_code_size;
-  codemask = 0;
   for (code = 0; code < clear_code; code++) {
     gfc->prefix[code] = 49428;
-    gfc->suffix[code] = code;
+    gfc->suffix[code] = (byte)code;
     gfc->length[code] = 1;
   }
   eoi_code = clear_code + 1;
   
-  next_code = eoi_code + 1;
-  max_code = clear_code << 1;
-  codemask = clear_code | (clear_code - 1);
+  next_code = eoi_code;
   bits_needed = min_code_size + 1;
-  amountleftover = 0;
-  leftover = 0;
-  GIF_DEBUG(("clear(%d) mask(%d) eoi(%d) needed(%d)", clear_code, codemask,
-	     eoi_code, bits_needed));
   
   code = clear_code;
   
-  buf = bufmax = buffer;
-  /* Thus buf == bufmax, and the `Read in the next data block.' code below
-     will be invoked on the first time through: exactly right! */
+  bit_length = bit_position = 0;
+  /* Thus the `Read in the next data block.' code below will be invoked on the
+     first time through: exactly right! */
   
   while (1) {
     
     old_code = code;
     
-    /* Get a code into the `code' variable. */
-    {
-      int had = amountleftover;
-      int left = bits_needed - amountleftover;
-      int shift = amountleftover;
-      
-      code = leftover;
-      
-      while (left > 0) {
-	if (buf == bufmax) {
-	  /* Read in the next data block. */
-	  i = gifgetbyte(grr);
-	  GIF_DEBUG(("imageblock(%d)", i));
-	  if (i == 0) goto zero_length_block;
-	  buf = buffer;
-	  bufmax = buffer + i;
-	  gifgetblock(buffer, i, grr);
-	}
-	leftover = *buf++;
-	code |= (leftover << shift);
-	shift += 8;
-	left -= 8;
-	had = 8;
-      }
-      
-      code &= codemask;
-      
-      if (left < 0) {
-	amountleftover = -left;
-	leftover >>= had + left;
-      } else {
-	leftover = 0;
-	amountleftover = 0;
-      }
-    }
+    /* GET A CODE INTO THE `code' VARIABLE.
+     * 
+     * 9.Dec.1998 - Rather than maintain a byte pointer and a bit offset into
+     * the current byte (and the processing associated with that), we maintain
+     * one number: the offset, in bits, from the beginning of `buffer'. This
+     * much cleaner choice was inspired by Patrick J. Naughton
+     * <naughton@wind.sun.com>'s GIF-reading code, which does the same thing.
+     * His code distributed as part of XV in xvgif.c. */
     
+    if (bit_position + bits_needed > bit_length)
+      /* Read in the next data block. */
+      if (!read_image_block(grr, buffer, &bit_position, &bit_length,
+			    bits_needed))
+	goto zero_length_block;
+    
+    i = bit_position / 8;
+    accum = buffer[i] + (buffer[i+1] << 8);
+    if (bits_needed >= 8)
+      accum |= (buffer[i+2]) << 16;
+    code = (Gif_Code)((accum >> (bit_position % 8)) & CUR_CODE_MASK);
+    bit_position += bits_needed;
+    
+    GIF_DEBUG(("code(%d)", code));
+    
+    /* CHECK FOR SPECIAL OR BAD CODES: clear_code, eoi_code, or a code that is
+     * too large. */
     if (code == clear_code) {
-      GIF_DEBUG(("!!"));
       bits_needed = min_code_size + 1;
-      next_code = eoi_code + 1;
-      max_code = clear_code << 1;
-      codemask = clear_code | (clear_code - 1);
+      next_code = eoi_code;
       continue;
       
     } else if (code == eoi_code)
       break;
-
+    
     else if (code > next_code && next_code) {
       /* code > next_code: a (hopefully recoverable) error.
 	 
-	 Bug fix, 5/27: Do this earlier, even if old_code == clear_code, and
-	 set code to 0 to prevent errors later. (If we didn't clear code, we'd
-	 later set old_code = code; then we had old_code >= next_code; so the
-	 prefixes array got all screwed up!) */
-      gfc->stream->errors++;
+	 Bug fix, 5/27: Do this even if old_code == clear_code, and set code
+	 to 0 to prevent errors later. (If we didn't zero code, we'd later set
+	 old_code = code; then we had old_code >= next_code; so the prefixes
+	 array got all screwed up!) */
+      gif_read_error(gfc, "unexpected code");
       code = 0;
     }
     
-    if (old_code != clear_code && next_code) {
-      /* Process the current code and define the next code. */
-      
-      /* *First,* set up the prefix and length for the next code
-	 (in case code == next_code). */
-      gfc->prefix[next_code] = old_code;
-      gfc->length[next_code] = gfc->length[old_code] + 1;
-      
-      /* Use one_code to process code. It's nice that it returns the first
-	 pixel in code: that's what we need. */
-      gfc->suffix[next_code] = one_code(gfc, code);
-      
-      /* Special processing if code == next_code: we didn't know code's final
-	 suffix when we called one_code, but we do now. */
-      if (code == next_code)
-	gfc->image[gfc->decodepos - 1] = gfc->suffix[next_code];
-      
-      next_code++;
-      
-    } else
-      /* Just process the current code. */
-      one_code(gfc, code);
+    /* PROCESS THE CURRENT CODE and define the next code. If no meaningful
+     * next code should be defined, then we have set next_code to either
+     * `eoi_code' or `clear_code' -- so we'll store useless prefix/suffix data
+     * in a useless place. */
     
-    if (next_code == max_code) {
-      if (bits_needed < GIF_MAX_CODE_SIZE) {
-	bits_needed++;
-	max_code <<= 1;
-	codemask = (codemask << 1) | 1;
-      } else
-	next_code = 0;
-      GIF_DEBUG(("%d/%d", bits_needed, max_code));
+    /* *First,* set up the prefix and length for the next code
+       (in case code == next_code). */
+    gfc->prefix[next_code] = old_code;
+    gfc->length[next_code] = gfc->length[old_code] + 1;
+    
+    /* Use one_code to process code. It's nice that it returns the first
+       pixel in code: that's what we need. */
+    gfc->suffix[next_code] = one_code(gfc, code);
+    
+    /* Special processing if code == next_code: we didn't know code's final
+       suffix when we called one_code, but we do now. */
+    if (code == next_code)
+      gfc->image[gfc->decodepos - 1] = gfc->suffix[next_code];
+    
+    /* Increment next_code except for the `clear_code' special case (that's
+       when we're reading at the end of a GIF) */
+    if (next_code != clear_code) {
+      next_code++;
+      if (next_code == CUR_BUMP_CODE) {
+	if (bits_needed < GIF_MAX_CODE_BITS)
+	  bits_needed++;
+	else
+	  next_code = clear_code;
+      }
     }
     
   }
   
   /* read blocks until zero-length reached. */
   i = gifgetbyte(grr);
+  GIF_DEBUG(("after_image(%d)", i));
   while (i > 0) {
     gifgetblock(buffer, i, grr);
     i = gifgetbyte(grr);
+    GIF_DEBUG(("after_image(%d)", i));
   }
   
   /* zero-length block reached. */
  zero_length_block:
   
-  if (gfc->image + gfc->decodepos != gfc->maximage)
-    /* Not enough data, or too much data, was read. */
-    gfc->stream->errors++;
+  if (gfc->image + gfc->decodepos < gfc->maximage)
+    gif_read_error(gfc, "not enough image data for image size");
+  else if (gfc->image + gfc->decodepos > gfc->maximage)
+    gif_read_error(gfc, "too much image data for image size");
 }
 
 
-static Gif_Color *
-color_table(int size, Gif_Reader *grr)
+static Gif_Colormap *
+read_color_table(int size, Gif_Reader *grr)
 {
+  Gif_Colormap *gfcm = Gif_NewFullColormap(size, size);
   Gif_Color *c;
-  Gif_Color *ct = Gif_NewArray(Gif_Color, size);
-  if (!ct) return 0;
+  if (!gfcm) return 0;
   
   GIF_DEBUG(("colormap(%d)", size));
-  for (c = ct; size; size--, c++) {
+  for (c = gfcm->col; size; size--, c++) {
     c->red = gifgetbyte(grr);
     c->green = gifgetbyte(grr);
     c->blue = gifgetbyte(grr);
     c->haspixel = 0;
   }
   
-  return ct;
+  return gfcm;
 }
 
 
 static int
-logical_screen_descriptor(Gif_Stream *gfs, Gif_Reader *grr)
+read_logical_screen_descriptor(Gif_Stream *gfs, Gif_Reader *grr)
      /* returns 0 on memory error */
 {
   byte packed;
@@ -371,11 +389,10 @@ logical_screen_descriptor(Gif_Stream *gfs, Gif_Reader *grr)
   gifgetbyte(grr);
   
   if (packed & 0x80) { /* have a global color table */
-    gfs->global = Gif_NewColormap();
+    int ncol = 1 << ((packed & 0x07) + 1);
+    gfs->global = read_color_table(ncol, grr);
     if (!gfs->global) return 0;
-    gfs->global->ncol = 1 << ((packed & 0x07) + 1);
-    gfs->global->col = color_table(gfs->global->ncol, grr);
-    if (!gfs->global->col) return 0;
+    gfs->global->refcount = 1;
   }
   
   return 1;
@@ -385,24 +402,20 @@ logical_screen_descriptor(Gif_Stream *gfs, Gif_Reader *grr)
 static int
 read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 {
-  byte buffer[GIF_MAX_BLOCK];
-  int i;
-  
   if (grr->is_record) {
     byte *first = grr->v;
-    byte *last;
-
-    /* min code size */
-    gifgetbyte(grr);
+    u_int32_t pos;
     
-    i = gifgetbyte(grr);
-    while (i > 0) {
-      gifgetblock(buffer, i, grr);
-      i = gifgetbyte(grr);
+    /* scan over image */
+    pos = 1;			/* skip min code size */
+    while (pos < grr->w) {
+      int amt = grr->v[pos];
+      pos += amt + 1;
+      if (amt == 0) break;
     }
-    last = grr->v;
+    if (pos > grr->w) pos = grr->w;
     
-    gfi->compressed_len = last - first;
+    gfi->compressed_len = pos;
     if (read_flags & GIF_READ_CONST_RECORD) {
       gfi->compressed = first;
       gfi->free_compressed = 0;
@@ -413,6 +426,10 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
       memcpy(gfi->compressed, first, gfi->compressed_len);
     }
     
+    /* move reader over that image */
+    grr->v += pos;
+    grr->w -= pos;
+    
   } else {
     /* non-record; have to read it block by block. */
     u_int32_t comp_cap = 1024;
@@ -454,23 +471,18 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 static int
 uncompress_image(Gif_Context *gfc, Gif_Image *gfi, Gif_Reader *grr)
 {
-  GIF_DEBUG(("%dx%d", gfi->width, gfi->height));
-  gfi->image_data = Gif_NewArray(byte, gfi->width * gfi->height);
-  gfi->free_image_data = Gif_DeleteArrayFunc;
-  if (!gfi->image_data) return 0;
-  if (!Gif_MakeImg(gfi, gfi->image_data, gfi->interlace)) return 0;
-  
+  if (!Gif_CreateUncompressedImage(gfi)) return 0;
   gfc->width = gfi->width;
   gfc->height = gfi->height;
   gfc->image = gfi->image_data;
   gfc->maximage = gfi->image_data + gfi->width * gfi->height;
-  image_data(gfc, grr);
+  read_image_data(gfc, grr);
   return 1;
 }
 
 
 int
-Gif_UncompressImage(Gif_Image *gfi)
+Gif_FullUncompressImage(Gif_Image *gfi, Gif_ReadErrorHandler h, void *hthunk)
 {
   Gif_Context gfc;
   Gif_Stream fake_gfs;
@@ -488,9 +500,11 @@ Gif_UncompressImage(Gif_Image *gfi)
   
   fake_gfs.errors = 0;
   gfc.stream = &fake_gfs;
-  gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE + 1);
-  gfc.suffix = Gif_NewArray(byte, GIF_MAX_CODE + 1);
-  gfc.length = Gif_NewArray(u_int16_t, GIF_MAX_CODE + 1);
+  gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE);
+  gfc.suffix = Gif_NewArray(byte, GIF_MAX_CODE);
+  gfc.length = Gif_NewArray(u_int16_t, GIF_MAX_CODE);
+  gfc.handler = h;
+  gfc.handler_thunk = hthunk;
   
   if (gfi && gfc.prefix && gfc.suffix && gfc.length && gfi->compressed) {
     make_data_reader(&grr, gfi->compressed, gfi->compressed_len);
@@ -505,7 +519,7 @@ Gif_UncompressImage(Gif_Image *gfi)
 
 
 static int
-gif_image(Gif_Reader *grr, Gif_Context *gfc, Gif_Image *gfi, int read_flags)
+read_image(Gif_Reader *grr, Gif_Context *gfc, Gif_Image *gfi, int read_flags)
      /* returns 0 on memory error */
 {
   byte packed;
@@ -515,14 +529,13 @@ gif_image(Gif_Reader *grr, Gif_Context *gfc, Gif_Image *gfi, int read_flags)
   gfi->width = gifgetunsigned(grr);
   gfi->height = gifgetunsigned(grr);
   packed = gifgetbyte(grr);
-  GIF_DEBUG(("< %ux%u > ", gfi->width, gfi->height));
+  GIF_DEBUG(("<%ux%u>", gfi->width, gfi->height));
   
   if (packed & 0x80) { /* have a local color table */
-    gfi->local = Gif_NewColormap();
+    int ncol = 1 << ((packed & 0x07) + 1);
+    gfi->local = read_color_table(ncol, grr);
     if (!gfi->local) return 0;
-    gfi->local->ncol = 1 << ((packed & 0x07) + 1);
-    gfi->local->col = color_table(gfi->local->ncol, grr);
-    if (!gfi->local->col) return 0;
+    gfi->local->refcount = 1;
   }
   
   gfi->interlace = (packed & 0x40) != 0;
@@ -557,7 +570,8 @@ gif_image(Gif_Reader *grr, Gif_Context *gfc, Gif_Image *gfi, int read_flags)
 
 
 static void
-graphic_control_extension(Gif_Stream *gfs, Gif_Image *gfi, Gif_Reader *grr)
+read_graphic_control_extension(Gif_Context *gfc, Gif_Image *gfi,
+			       Gif_Reader *grr)
 {
   byte len;
   byte crap[GIF_MAX_BLOCK];
@@ -575,13 +589,13 @@ graphic_control_extension(Gif_Stream *gfs, Gif_Image *gfi, Gif_Reader *grr)
   }
   
   if (len > 0) {
-    gfs->errors++;
+    gif_read_error(gfc, "odd graphic extension format");
     gifgetblock(crap, len, grr);
   }
   
   len = gifgetbyte(grr);
   while (len > 0) {
-    gfs->errors++;
+    gif_read_error(gfc, "odd graphic extension format");
     gifgetblock(crap, len, grr);
     len = gifgetbyte(grr);
   }
@@ -614,8 +628,8 @@ suck_data(char *data, int *store_len, Gif_Reader *grr)
 
 
 static int
-unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
-		  Gif_Reader *grr)
+read_unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
+		       Gif_Reader *grr)
 {
   byte block_len = gifgetbyte(grr);
   byte *data = 0;
@@ -623,18 +637,20 @@ unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
   Gif_Extension *gfex = 0;
 
   while (block_len > 0) {
-    if (data) Gif_ReArray(data, byte, data_len + block_len);
-    else data = Gif_NewArray(byte, block_len);
+    if (data) Gif_ReArray(data, byte, data_len + block_len + 1);
+    else data = Gif_NewArray(byte, block_len + 1);
     if (!data) goto done;
     gifgetblock(data + data_len, block_len, grr);
     data_len += block_len;
     block_len = gifgetbyte(grr);
   }
-  
-  gfex = Gif_NewExtension(kind, app_name);
+
+  if (data)
+    gfex = Gif_NewExtension(kind, app_name);
   if (gfex) {
     gfex->data = data;
     gfex->length = data_len;
+    data[data_len] = 0;
     Gif_AddExtension(gfs, gfex, position);
   }
   
@@ -650,9 +666,10 @@ unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
 
 
 static int
-application_extension(Gif_Stream *gfs, Gif_Image *gfi, int position,
-		      Gif_Reader *grr)
+read_application_extension(Gif_Context *gfc, Gif_Image *gfi, int position,
+			   Gif_Reader *grr)
 {
+  Gif_Stream *gfs = gfc->stream;
   byte buffer[GIF_MAX_BLOCK + 1];
   byte len = gifgetbyte(grr);
   gifgetblock(buffer, len, grr);
@@ -665,9 +682,9 @@ application_extension(Gif_Stream *gfs, Gif_Image *gfi, int position,
       gifgetbyte(grr); /* throw away the 1 */
       gfs->loopcount = gifgetunsigned(grr);
       len = gifgetbyte(grr);
-      if (len) gfs->errors++;
+      if (len) gif_read_error(gfc, "bad loop extension");
     } else
-      gfs->errors++;
+      gif_read_error(gfc, "bad loop extension");
     
     while (len > 0) {
       gifgetblock(buffer, len, grr);
@@ -677,13 +694,13 @@ application_extension(Gif_Stream *gfs, Gif_Image *gfi, int position,
     
   } else {
     buffer[len] = 0;
-    return unknown_extension(gfs, 0xFF, (char *)buffer, position, grr);
+    return read_unknown_extension(gfs, 0xFF, (char *)buffer, position, grr);
   }
 }
 
 
 static int
-comment_extension(Gif_Image *gfi, Gif_Reader *grr)
+read_comment_extension(Gif_Image *gfi, Gif_Reader *grr)
 {
   int len;
   Gif_Comment *gfcom = gfi->comment;
@@ -699,7 +716,8 @@ comment_extension(Gif_Image *gfi, Gif_Reader *grr)
 
 
 static Gif_Stream *
-gif_read(Gif_Reader *grr, int read_flags)
+read_gif(Gif_Reader *grr, int read_flags,
+	 Gif_ReadErrorHandler handler, void *handler_thunk)
 {
   Gif_Stream *gfs;
   Gif_Image *gfi;
@@ -720,22 +738,23 @@ gif_read(Gif_Reader *grr, int read_flags)
   gfi = Gif_NewImage();
   
   gfc.stream = gfs;
-  gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE + 1);
-  gfc.suffix = Gif_NewArray(byte, GIF_MAX_CODE + 1);
-  gfc.length = Gif_NewArray(u_int16_t, GIF_MAX_CODE + 1);
+  gfc.prefix = Gif_NewArray(Gif_Code, GIF_MAX_CODE);
+  gfc.suffix = Gif_NewArray(byte, GIF_MAX_CODE);
+  gfc.length = Gif_NewArray(u_int16_t, GIF_MAX_CODE);
+  gfc.handler = handler;
+  gfc.handler_thunk = handler_thunk;
   
   if (!gfs || !gfi || !gfc.prefix || !gfc.suffix || !gfc.length)
     goto done;
   
-  GIF_DEBUG(("GIF"));
-  if (!logical_screen_descriptor(gfs, grr))
+  GIF_DEBUG(("\nGIF"));
+  if (!read_logical_screen_descriptor(gfs, grr))
     goto done;
   GIF_DEBUG(("logscrdesc"));
   
   while (!gifeof(grr)) {
     
     byte block = gifgetbyte(grr);
-    GIF_DEBUG(("block(%x)", block));
     
     switch (block) {
       
@@ -744,7 +763,8 @@ gif_read(Gif_Reader *grr, int read_flags)
       
       gfi->identifier = last_name;
       last_name = 0;
-      if (!gif_image(grr, &gfc, gfi, read_flags) || !Gif_AddImage(gfs, gfi)) {
+      if (!read_image(grr, &gfc, gfi, read_flags)
+	  || !Gif_AddImage(gfs, gfi)) {
 	Gif_DeleteImage(gfi);
 	goto done;
       }
@@ -762,17 +782,17 @@ gif_read(Gif_Reader *grr, int read_flags)
       break;
       
      case ';': /* terminator */
-      GIF_DEBUG(("fuck"));
+      GIF_DEBUG(("term"));
       ok = 1;
       goto done;
       
      case '!': /* extension */
-      GIF_DEBUG(("ext"));
       block = gifgetbyte(grr);
+      GIF_DEBUG(("ext(0x%02X)", block));
       switch (block) {
 	
        case 0xF9:
-	graphic_control_extension(gfs, gfi, grr);
+	read_graphic_control_extension(&gfc, gfi, grr);
 	break;
 	
        case 0xCE:
@@ -780,22 +800,22 @@ gif_read(Gif_Reader *grr, int read_flags)
 	break;
 	
        case 0xFE:
-	if (!comment_extension(gfi, grr)) goto done;
+	if (!read_comment_extension(gfi, grr)) goto done;
 	break;
 	
        case 0xFF:
-	application_extension(gfs, gfi, extension_position, grr);
+	read_application_extension(&gfc, gfi, extension_position, grr);
 	break;
 	
        default:
-	unknown_extension(gfs, block, 0, extension_position, grr);
+	read_unknown_extension(gfs, block, 0, extension_position, grr);
 	break;
 
       }
       break;
       
      default:
-      gfs->errors++;
+      gif_read_error(&gfc, "unknown block type");
       break;
       
     }
@@ -823,26 +843,29 @@ gif_read(Gif_Reader *grr, int read_flags)
 
 
 Gif_Stream *
-Gif_FullReadFile(FILE *f, int read_flags)
+Gif_FullReadFile(FILE *f, int read_flags,
+		 Gif_ReadErrorHandler h, void *hthunk)
 {
   Gif_Reader grr;
+  if (!f) return 0;
   grr.f = f;
   grr.is_record = 0;
   grr.byte_getter = file_byte_getter;
   grr.block_getter = file_block_getter;
   grr.eofer = file_eofer;
-  return gif_read(&grr, read_flags);
+  return read_gif(&grr, read_flags, h, hthunk);
 }
 
-
 Gif_Stream *
-Gif_FullReadRecord(Gif_Record *gifrec, int read_flags)
+Gif_FullReadRecord(Gif_Record *gifrec, int read_flags,
+		   Gif_ReadErrorHandler h, void *hthunk)
 {
   Gif_Reader grr;
+  if (!gifrec) return 0;
   make_data_reader(&grr, gifrec->data, gifrec->length);
   if (read_flags & GIF_READ_CONST_RECORD)
     read_flags |= GIF_READ_COMPRESSED;
-  return gif_read(&grr, read_flags);
+  return read_gif(&grr, read_flags, h, hthunk);
 }
 
 
